/* XOSD

Copyright (c) 2006 Christian Birchinger (joker@netswarm.net)

Copyright (c) 2001 Andre Renaud (andre@ignavus.net)

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <ctype.h>
#include <gtk/gtk.h>

#include "audacious_osd.h"

#include <audacious/plugin.h>
#include <audacious/beepctrl.h>
#include <audacious/util.h>
#include <audacious/configdb.h>

static void init(void);
static void cleanup(void);
static gint timeout_func(gpointer);

GeneralPlugin gp = {
  .handle = NULL,
  .filename = NULL,
  .xmms_session = -1,
  .description = "On Screen Display " PLUGIN_VERSION,
  .init = init,
  .about = show_about,
  .configure = configure,
  .cleanup = cleanup,
};


struct state
{
  gboolean playing;
  gboolean paused;
  gboolean shuffle;
  gboolean repeat;
  int pos;
  int volume;
  int balance;
  gchar *title;
};
static struct state previous;
struct show show;
struct xosdconf xosdconf;

xosd *osd = NULL;
static guint timeout_tag;

/*
 * Return plugin structure.
 */
GeneralPlugin *
get_gplugin_info(void)
{
  return &gp;
}


/*
 * Initialize plugin.
 */
static void
init(void)
{
  DEBUG("init");

  if (osd) {
    DEBUG("uniniting osd");
    xosd_destroy(osd);
    osd = NULL;
  }

  read_config();

#if 0
  /* Deadlocks, because it is to early. */
  previous.playing = xmms_remote_is_playing(gp.xmms_session);
  previous.paused = xmms_remote_is_paused(gp.xmms_session);
  previous.shuffle = xmms_remote_is_shuffle(gp.xmms_session);
  previous.repeat = xmms_remote_is_repeat(gp.xmms_session);
  previous.pos = xmms_remote_get_playlist_pos(gp.xmms_session);
  previous.volume = xmms_remote_get_main_volume(gp.xmms_session);
  previous.balance = (xmms_remote_get_balance(gp.xmms_session) + 100) / 2;
  previous.title = NULL;
#else
  memset(&previous, 0, sizeof(struct state));
#endif

  DEBUG("calling osd init function");

  osd = xosd_create(2);
  apply_config();
  DEBUG("osd initialized");
  if (osd)
    timeout_tag = gtk_timeout_add(100, timeout_func, NULL);
}

/*
 * Free memory and release resources.
 */
static void
cleanup(void)
{
  DEBUG("cleanup");
  if (osd && timeout_tag)
    gtk_timeout_remove(timeout_tag);
  timeout_tag = 0;

  if (xosdconf.font) {
    g_free(xosdconf.font);
    xosdconf.font = NULL;
  }

  if (xosdconf.colour) {
    g_free(xosdconf.colour);
    xosdconf.colour = NULL;
  }

  if (previous.title) {
    g_free(previous.title);
    previous.title = NULL;
  }

  if (osd) {
    DEBUG("hide");
    xosd_hide(osd);
    DEBUG("uninit");
    xosd_destroy(osd);
    DEBUG("done with osd");
    osd = NULL;
  }
}

/*
 * Display about dialog.
 */
void
show_about(void)
{
  static GtkWidget *about;
  if (about) return;
  about = xmms_show_message("About XOSD Plugin",
                            "XOSD plugin " PLUGIN_VERSION "\n\n"
                            "Audcious port:\n"
                            "Copyright (c) 2006 Christian Birchinger <joker@netswarm.net>\n\n"
                            "Original XMMS/BMP plugin:\n"
                            "Copyright (c) 2001 Andre Renaud <andre@ignavus.net>\n\n"
                            "Built with XOSD version " XOSD_VERSION,
                            "Ok", FALSE, NULL, NULL);

  g_signal_connect(G_OBJECT(about), "destroy",
                   G_CALLBACK(gtk_widget_destroyed), &about);
}

/*
 * Read configuration and initialize variables.
 */
void
read_config(void)
{
  ConfigDb *db;
  DEBUG("read_config");

  show.volume = 1;
  show.balance = 1;
  show.pause = 1;
  show.trackname = 1;
  show.stop = 1;
  show.repeat = 1;
  show.shuffle = 1;
  show.listpos = 1;

  g_free(xosdconf.colour);
  g_free(xosdconf.font);
  xosdconf.colour = NULL;
  xosdconf.font = NULL;
  xosdconf.timeout = 3;
  xosdconf.offset = 50;
  xosdconf.h_offset = 0;
  xosdconf.shadow_offset = 1;
  xosdconf.outline_offset = 0;
  xosdconf.pos = XOSD_bottom;
  xosdconf.align = XOSD_left;

  DEBUG("reading configuration data");
  if ((db = bmp_cfg_db_open()) != NULL) {
    bmp_cfg_db_get_string(db, "XOSD", "font", &xosdconf.font);
    bmp_cfg_db_get_string(db, "XOSD", "colour", &xosdconf.colour);
    bmp_cfg_db_get_int(db, "XOSD", "timeout", &xosdconf.timeout);
    bmp_cfg_db_get_int(db, "XOSD", "offset", &xosdconf.offset);
    bmp_cfg_db_get_int(db, "XOSD", "h_offset", &xosdconf.h_offset);
    bmp_cfg_db_get_int(db, "XOSD", "shadow_offset", &xosdconf.shadow_offset);
    bmp_cfg_db_get_int(db, "XOSD", "outline_offset", &xosdconf.outline_offset);
    bmp_cfg_db_get_int(db, "XOSD", "pos", &xosdconf.pos);
    bmp_cfg_db_get_int(db, "XOSD", "align", &xosdconf.align);
    bmp_cfg_db_get_int(db, "XOSD", "show_volume", &show.volume);
    bmp_cfg_db_get_int(db, "XOSD", "show_balance", &show.balance);
    bmp_cfg_db_get_int(db, "XOSD", "show_pause", &show.pause);
    bmp_cfg_db_get_int(db, "XOSD", "show_trackname", &show.trackname);
    bmp_cfg_db_get_int(db, "XOSD", "show_stop", &show.stop);
    bmp_cfg_db_get_int(db, "XOSD", "show_repeat", &show.repeat);
    bmp_cfg_db_get_int(db, "XOSD", "show_shuffle", &show.shuffle);
    bmp_cfg_db_get_int(db, "XOSD", "show_listpos", &show.listpos);
    bmp_cfg_db_close(db);
  }

  DEBUG("getting default font");
  if (xosdconf.font == NULL)
    xosdconf.font = g_strdup(osd_default_font);

  DEBUG("default colour");
  if (xosdconf.colour == NULL)
    xosdconf.colour = g_strdup("green");

  DEBUG("done");
}

/*
 * Write configuration.
 */
void
write_config(void)
{
  ConfigDb *db;
  DEBUG("write_config");

  if ((db = bmp_cfg_db_open()) != NULL) {
    bmp_cfg_db_set_string(db, "XOSD", "colour", xosdconf.colour);
    bmp_cfg_db_set_string(db, "XOSD", "font", xosdconf.font);
    bmp_cfg_db_set_int(db, "XOSD", "timeout", xosdconf.timeout);
    bmp_cfg_db_set_int(db, "XOSD", "offset", xosdconf.offset);
    bmp_cfg_db_set_int(db, "XOSD", "h_offset", xosdconf.h_offset);
    bmp_cfg_db_set_int(db, "XOSD", "shadow_offset", xosdconf.shadow_offset);
    bmp_cfg_db_set_int(db, "XOSD", "outline_offset", xosdconf.outline_offset);
    bmp_cfg_db_set_int(db, "XOSD", "pos", xosdconf.pos);
    bmp_cfg_db_set_int(db, "XOSD", "align", xosdconf.align);

    bmp_cfg_db_set_int(db, "XOSD", "show_volume", show.volume);
    bmp_cfg_db_set_int(db, "XOSD", "show_balance", show.balance);
    bmp_cfg_db_set_int(db, "XOSD", "show_pause", show.pause);
    bmp_cfg_db_set_int(db, "XOSD", "show_trackname", show.trackname);
    bmp_cfg_db_set_int(db, "XOSD", "show_stop", show.stop);
    bmp_cfg_db_set_int(db, "XOSD", "show_repeat", show.repeat);
    bmp_cfg_db_set_int(db, "XOSD", "show_shuffle", show.shuffle);
    bmp_cfg_db_set_int(db, "XOSD", "show_listpos", show.listpos);

    bmp_cfg_db_close(db);
  }

  DEBUG("done");
}

void
apply_config(void)
{
  DEBUG("apply_config");
  if (osd) {
    if (xosd_set_font(osd, xosdconf.font) == -1)
      DEBUG("invalid font %s", xosdconf.font);

    xosd_set_colour(osd, xosdconf.colour);
    xosd_set_timeout(osd, xosdconf.timeout);
    xosd_set_shadow_offset(osd, xosdconf.shadow_offset);
    xosd_set_outline_offset(osd, xosdconf.outline_offset);
    xosd_set_pos(osd, xosdconf.pos);
    xosd_set_align(osd, xosdconf.align);
    xosd_set_vertical_offset(osd, xosdconf.offset);
    xosd_set_horizontal_offset(osd, xosdconf.h_offset);
  }
  DEBUG("done");
}

/*
 * Convert hexcode to ASCII.
 */
static void
replace_hexcodes(gchar * text)
{
  gchar *head, *tail;
  int conv_underscore, c;
  ConfigDb *db;
  DEBUG("replace_hexcodes");

  if ((db = bmp_cfg_db_open()) != NULL) {
    bmp_cfg_db_get_bool(db, NULL, "convert_underscore",
                          &conv_underscore);
    bmp_cfg_db_close(db);
  }

  for (head = tail = text; *head; head++, tail++) {
    /* replace underscors with spaces if necessary */
    if (conv_underscore && *head == '_') {
      *tail = ' ';
      continue;
    }
    /* replace hex with character if necessary */
    if (*head == '%' && isxdigit(head[1]) && isxdigit(head[2])) {
      sscanf(head + 1, "%2x", &c);
      *tail = (char) c;
      head += 2;
      continue;
    }
    *tail = *head;
  }
  *tail = '\0';
}

/*
 * Callback funtion to handle delayed display.
 */
static gint
timeout_func(gpointer data)
{
  struct state current;
  char *text = NULL;
  gboolean songchange, showtext, withtime = FALSE;
  gint playlist_length;

  DEBUG("timeout func");

  if (!osd)
    return FALSE;

  GDK_THREADS_ENTER();

  current.playing = xmms_remote_is_playing(gp.xmms_session);
  current.paused = xmms_remote_is_paused(gp.xmms_session);
  current.shuffle = xmms_remote_is_shuffle(gp.xmms_session);
  current.repeat = xmms_remote_is_repeat(gp.xmms_session);
  current.pos = xmms_remote_get_playlist_pos(gp.xmms_session);
  current.volume = xmms_remote_get_main_volume(gp.xmms_session);
  current.balance = (xmms_remote_get_balance(gp.xmms_session) + 100) / 2;

  /* Get the current title only if the playlist is not empty. Otherwise
   * it'll crash. Don't forget to free the title! */
  playlist_length = xmms_remote_get_playlist_length(gp.xmms_session);
  current.title = playlist_length
    ? xmms_remote_get_playlist_title(gp.xmms_session, current.pos) : NULL;

  /* Check for song change. Deleting a song from the playlist only changed the
   * name, but not the position, so compare also by (still hexencoded) name. */
  songchange =
    (previous.pos != current.pos) ||
    (previous.title == NULL && current.title != NULL) ||
    (previous.title != NULL && current.title == NULL) ||
    (previous.title != NULL && current.title != NULL &&
     (g_strcasecmp(previous.title, current.title) != 0));

  /* Possible show something when either song or state changed. */
  showtext = songchange ||
    ((current.playing != previous.playing) ||
     (current.paused != previous.paused));

  /* Determine right text depending on state and state/title change.
   *    +---+          +---+
   *    |   |          |   |
   *    +->PLAY<---->STOP<-+
   *        ^          ^
   *        |          |
   *        +-->PAUSE--+
   */
  if (!current.playing) {       /* {PLAY,PAUSE,STOP} -> STOP */
    text = "Stopped";
    showtext &= show.stop;
  } else if (current.paused) {  /* PLAY -> PAUSE */
    text = "Paused";
    showtext &= show.pause;
    withtime = TRUE;
  } else if (previous.paused && !current.paused && !songchange) {       /* PAUSE -SameSong-> PLAY */
    text = "Unpaused";
    showtext &= show.pause;
    withtime = TRUE;
  } else {                      /* {PLAY,STOP} -> PLAY <-OtherSong- PAUSE */
    text = "Playing";
    showtext &= show.trackname;
  }

  /* Decide what to display, in decreasing priority. */
  if (showtext) {
    xosd_display(osd, 0, XOSD_string, text);
    if (show.trackname && (current.title != NULL)) {
      int len;
      char *title;
      gint playlist_time;

      len = 13 + strlen(current.title) + (withtime ? 11 : 0);
      title = malloc(len);
      playlist_time =
        withtime ? xmms_remote_get_output_time(gp.xmms_session) : 0;
      if (show.listpos)
        snprintf(title, len,
               withtime ? "%i/%i: %s (%i:%02i)" : "%i/%i: %s",
               current.pos + 1, playlist_length, current.title,
               playlist_time / 1000 / 60, playlist_time / 1000 % 60);
      else
        snprintf(title, len,
               withtime ? "%s (%i:%02i)" : "%s",
               current.title,
               playlist_time / 1000 / 60, playlist_time / 1000 % 60);

      replace_hexcodes(title);
      xosd_display(osd, 1, XOSD_string, title);
      free(title);
    } else
      xosd_display(osd, 1, XOSD_string, "");
  } else if (current.volume != previous.volume && show.volume) {
    DEBUG("V: %d->%d\n", previous.volume, current.volume);
    /* xmms returns -1 during a title change. skip this and try again later. */
    if ((previous.volume == -1) || (current.volume == -1))
      goto skip;
    xosd_display(osd, 0, XOSD_string, "Volume");
    xosd_display(osd, 1, XOSD_percentage, current.volume);
  } else if (current.balance != previous.balance && show.balance) {
    DEBUG("B: %d->%d\n", previous.balance, current.balance);
    /* FIXME: Same as above might happen, but with what values? */
    xosd_display(osd, 0, XOSD_string, "Balance");
    xosd_display(osd, 1, XOSD_slider, current.balance);
  } else if (current.repeat != previous.repeat && show.repeat) {
    xosd_display(osd, 0, XOSD_string, "Repeat");
    xosd_display(osd, 1, XOSD_string, current.repeat ? "On" : "Off");
  } else if (current.shuffle != previous.shuffle && show.shuffle) {
    xosd_display(osd, 0, XOSD_string, "Shuffle");
    xosd_display(osd, 1, XOSD_string, current.shuffle ? "On" : "Off");
  }

skip:
  /* copy current state (including title) for future comparison. Free old
   * title first. */
  if (previous.title)
    g_free(previous.title);
  previous = current;

  GDK_THREADS_LEAVE();


  return TRUE;
}

/* vim: tabstop=8 shiftwidth=8 noexpandtab
 */
